#include "image.h"
#include <math.h>
#include <limits.h>

#define REPORT 1

#ifdef IMAGE_CHECKALL
static char rcsid[] = "<motion> encountered an image check failure.";
#endif

/* Read image intensities from image. */
#define I(image, i, j) imRef(image, i, j)

/* Simplify calculation of a square. */
#define SQR(x) ((x) * (x))

/* Easy access to storage and retrieval from matrices. */
#define STORE_MEAN(i, j, mean) imRef(meanMatrix, i, j) = mean
#define STORE_DEV(i, j, dev) imRef(stdDevMatrix, i, j) = dev
#define MEAN(i, j) imRef(meanMatrix, i, j)
#define STD_DEV(i, j) imRef(stdDevMatrix, i, j)

/* Global variable declarations. */

GrayImage imageA, imageB; /* Input images. */
ShortImage dxDisparityImage, dyDisparityImage; /* Output disparity images. */

char *imageAFileName; 
char *imageBFileName;                     /* Command line arguments. */
int hMaxDisparity, vMaxDisparity;
int hConvNeighborhood, vConvNeighborhood;
int hRankNeighborhood = 0;
int vRankNeighborhood = 0;

/* Function prototypes. */
void normCorrelation(int imageWidth, int imageHeight);
void rankTransform(int imageWidth, int imageHeight);
float bruteCalcMatchWindowMean(GrayImage image,
			       int matchWindowTop, int matchWindowBottom,
			       int matchWindowLeft, int matchWindowRight);
float bruteCalcMatchWindowStdDev(GrayImage image, float matchWindowMean,
				 int matchWindowTop, int matchWindowBottom,
				 int matchWindowLeft, int matchWindowRight);
void writeDisparityImageFile(ShortImage image, char *filename,
			     int hDisparity, int vDisparity,
			     int hNeighborhood, int vNeighborhood);
GrayImage createRankTransformedImage(GrayImage image, int imageWidth,
				     int ImageHeight);
/* Main program starts here. */

int main(int argc, char **argv) {
  /* Declaration of variables local to _main_. */
  
  int imageWidth, imageHeight;
  int croppedWidth, croppedHeight;
  int hOffset, vOffset;
  char *filename;
  
  /* Make sure we have enough input arguments. */
  if (argc < 6) {
    /* motion <imageA> <imageB> <hdisp> <vdisp> <hsearch> <vsearch>
                                                             [hrank] [vrank] */
    printf("Too few arguments given to program. Aborting.\n");
    exit(1);
  }

  /* Read in the parameters from the command line. */
  imageAFileName = argv[1]; /* Filenames for images A and B. */
  imageBFileName = argv[2];
  hMaxDisparity = atoi(argv[3]); /* Maximum horizontal and vertical */
  vMaxDisparity = atoi(argv[4]); /* disparities. */
  hConvNeighborhood = atoi(argv[5]); /* Dimensions of convolution */
  vConvNeighborhood = atoi(argv[6]); /* neighborhood. */

  /* If the user has given additional arguments, the dimensions of the rank
     neighborhood, we do rank transform, otherwise we use normalized
     correlation. */
  
  if (argc > 6) {
    hRankNeighborhood = atoi(argv[7]);
    vRankNeighborhood = atoi(argv[8]);
  }
  
  /* Let the user know what the program _thinks_ it's doing. */
  printf("\nRUNNING: motion %s %s %d %d %d %d",
	 imageAFileName, imageBFileName, hMaxDisparity,
	 vMaxDisparity, hConvNeighborhood, vConvNeighborhood);
  if (hRankNeighborhood > 0) {
    printf(" %d %d", hRankNeighborhood, vRankNeighborhood);
  }
  printf("\n\n");
  
#ifdef REPORT
  /* Print debugging info about the command line arguments. */
  printf("Max horizontal disparity: %d\nMax vertical disparity: %d\n",
	 hMaxDisparity, vMaxDisparity);
  printf("Conv neighborhood width: %d\nConv neighborhood height: %d\n\n",
	 hConvNeighborhood, vConvNeighborhood);
#endif
  
  /* Attempt to load in the input images, and check that attempt
     was successful. */
  imageA = (GrayImage)imLoad(IMAGE_GRAY, imageAFileName);
  imageB = (GrayImage)imLoad(IMAGE_GRAY, imageBFileName);
  if ((imageA == (GrayImage)NULL) || (imageB == (GrayImage)NULL)) {
    printf("Error loading one of the input images! Aborting.\n");
    exit(1);
  }
  
  /* Determine the dimensions of the input images, and do a sanity check. */
  imageWidth = (int)imGetWidth(imageA); 
  imageHeight = (int)imGetHeight(imageA);
  
  if (((int)imGetWidth(imageB) != imageWidth) ||
      ((int)imGetHeight(imageB) != imageHeight)) {
    printf("Image dimensions do not match! Aborting.\n");
    exit(1);
  }

  /* The area of the images for which motion disparities can be computed
     is a function of the size of the match and search windows, and for
     rank transform, the rank neighborhood.  */
  hOffset = hMaxDisparity + hConvNeighborhood + hRankNeighborhood;
  vOffset = vMaxDisparity + vConvNeighborhood + vRankNeighborhood;

  croppedWidth = imageWidth - 2*(hOffset);
  croppedHeight = imageHeight - 2*(vOffset);
  
#ifdef REPORT
  /* Print debugging info about image dimensions. */
  printf("%ux%u input images loaded.\n", imageWidth, imageHeight); 
  printf("Computable area = %ux%u\n", croppedWidth, croppedHeight);
  /*  printf("hOffset = %u\nvOffset = %d\n", hOffset, vOffset); */
#endif

  /* Create empty dx and dy motion disparity output images, checking
     for success. */
  dxDisparityImage =
    (ShortImage)imNew(IMAGE_SHORT, croppedWidth, croppedHeight);
  dyDisparityImage =
    (ShortImage)imNew(IMAGE_SHORT, croppedWidth, croppedHeight);

  imSetOffset(dxDisparityImage, hOffset, vOffset);
  imSetOffset(dyDisparityImage, hOffset, vOffset);
  
  if ((dxDisparityImage == (ShortImage)NULL) ||
      (dyDisparityImage == (ShortImage)NULL)) {
    printf("Problems creating output disparity images! Aborting.\n");
    exit(1);
  }

#ifdef REPORT
  printf("Motion disparity images created.\n\n");
#endif

  if (hRankNeighborhood > 0) {
    rankTransform(imageWidth, imageHeight);
  }
  else {
    normCorrelation(imageWidth, imageHeight);
  }
  
  /* Write motion disparity images to disk. */
  
  filename = (char *)malloc(25*sizeof(char));
  sprintf(filename, "dx_disparity.psm");
  
  writeDisparityImageFile(dxDisparityImage, filename,
			  hMaxDisparity, vMaxDisparity,
			  hConvNeighborhood, vConvNeighborhood);
#ifdef REPORT
  printf("Wrote %s.\n", filename);
#endif

  sprintf(filename, "dy_disparity.psm");

  writeDisparityImageFile(dyDisparityImage, filename, 
			  hMaxDisparity, vMaxDisparity,
			  hConvNeighborhood, vConvNeighborhood);
#ifdef REPORT
  printf("Wrote %s.\n\n", filename);
#endif
  printf("background = %d", imRef(dxDisparityImage, 20, 20));
  /* Clean up the images that we used. */
  imFree(imageA);
  printf("Removed imageA.\n");
  imFree(imageB);
  printf("Removed imageB.\n");
  imFree(dxDisparityImage);
  printf("Removed dxDisparityImage.\n");
  imFree(dyDisparityImage);
  printf("Removed dyDisparityImage.\n");

  /* Get outta here. */
  return(0);
}

/* normCorrelation
 *
 *     input: integer dimensions of the input images
 *     output: void
 *
 *    Computes horizontal and vertical motion disparity images using
 *    normalized correlation.  Function is executed for effect on global
 *    output images.
 **/

void normCorrelation(int imageWidth, int imageHeight) {
  
  /* Declaration of variables local to _normCorrelation_. */
  FloatImage meanMatrix, stdDevMatrix;

  int xdiff, ydiff;
  int dataMatrixWidth, dataMatrixHeight;
  int croppedImageTop, croppedImageBottom, croppedImageLeft, croppedImageRight;
  int matchWindowTop, matchWindowBottom, matchWindowLeft, matchWindowRight;
  int matchWindowWidth, matchWindowHeight, N;
  int searchWindowTop, searchWindowBottom, searchWindowLeft, searchWindowRight;
  int imageI, imageJ;
  int searchI, searchJ;
  int matchI, matchJ;
  short dx, dy;
  float imageAMean, imageAStdDev, imageBMean, imageBStdDev;
  float minR, currentR;
  float runningSum;

#ifdef REPORT
  printf("Entering normalized correlation algorithm...\n");
#endif

  
  /* Figure out the boundaries of the "cropped image," the area of the
     image for which motion disparities are computable. */
  croppedImageTop = (vMaxDisparity + vConvNeighborhood);
  croppedImageBottom = (imageHeight - (vMaxDisparity + vConvNeighborhood) - 1);
  croppedImageLeft = (hMaxDisparity + hConvNeighborhood);
  croppedImageRight = (imageWidth - (hMaxDisparity + hConvNeighborhood) - 1);
  
  /* Create matrix images to store match window means and standard
     deviations as they are dynamically computed. Then set the origin of
     the matrices so that pixels match with the input images. Finish up
     with a sanity check. */
  dataMatrixWidth = (imageWidth - 2*hConvNeighborhood);
  dataMatrixHeight = (imageHeight - 2*vConvNeighborhood);

  meanMatrix =
    (FloatImage)imNew(IMAGE_FLOAT, dataMatrixWidth, dataMatrixHeight);
  stdDevMatrix =
    (FloatImage)imNew(IMAGE_FLOAT, dataMatrixWidth, dataMatrixHeight);
  
  imSetOffset(meanMatrix, hConvNeighborhood, vConvNeighborhood);
  imSetOffset(stdDevMatrix, hConvNeighborhood, vConvNeighborhood);
  
  if ((meanMatrix == (FloatImage)NULL) ||
      (stdDevMatrix == (FloatImage)NULL)) {
    printf("Could not create either meanMatrix or stdDevMatrix! Aborting.\n");
    exit(1);
  }
  
  
  /* Compute the dimensions of the match window, and the number of
     pixels in the window. */
  matchWindowWidth = (2*hConvNeighborhood + 1);
  matchWindowHeight = (2*vConvNeighborhood + 1);
  N = matchWindowWidth * matchWindowHeight;
  
  /* Loop over the area of the image for which motion disparities are
     computable. */
  for (imageJ = croppedImageTop; imageJ <= croppedImageBottom; imageJ++) {

#ifdef REPORT
    printf("Evaluating disparities for pixels in row %d.\n", imageJ);
#endif

    for (imageI = croppedImageLeft; imageI <= croppedImageRight; imageI++) {


      /* Initally, the minimum r is "infinity" and disparities are zero. */
      minR = 0;
      dx = 0;
      dy = 0;
      
      /* Find the boundaries of the match window for pixel (imageI, imageJ). */
      matchWindowTop = (imageJ - vConvNeighborhood);
      matchWindowBottom = (imageJ + vConvNeighborhood);
      matchWindowLeft = (imageI - hConvNeighborhood);
      matchWindowRight = (imageI + hConvNeighborhood);
      
      /* Compute mean and std dev for match window in image A. */
      imageAMean =
	bruteCalcMatchWindowMean(imageA, matchWindowTop, matchWindowBottom,
				 matchWindowLeft, matchWindowRight);
      imageAStdDev =
	bruteCalcMatchWindowStdDev(imageA, imageAMean,
				   matchWindowTop, matchWindowBottom,
				   matchWindowLeft, matchWindowRight);
      
      /* Find the boundaries of the search window for pixel
	 (imageI, imageJ). */
      searchWindowTop = (imageJ - vMaxDisparity);
      searchWindowBottom = (imageJ + vMaxDisparity);
      searchWindowLeft = (imageI - hMaxDisparity);
      searchWindowRight = (imageI + hMaxDisparity);

      
      /* Iterate over these search window in image B. */
      for (searchJ=searchWindowTop;searchJ<= searchWindowBottom; searchJ++) {
	for (searchI=searchWindowLeft;searchI<= searchWindowRight; searchI++) {
	  
	  /* Find the boundaries of the match window for
	     the current location in the search window. */
	  matchWindowTop = (searchJ - vConvNeighborhood);
	  matchWindowBottom = (searchJ + vConvNeighborhood);
	  matchWindowLeft = (searchI - hConvNeighborhood);
	  matchWindowRight = (searchI + hConvNeighborhood);
	  
	  /* If we've been to this pixel before, we can save by grabbing
	     the mean and std dev that we stored, otherwise we've got
	     to compute them by brute force. */
	  imageBMean = MEAN(searchI, searchJ);
	  imageBStdDev = STD_DEV(searchI, searchJ);
	  if ((imageBMean == 0) && (imageBStdDev == 0)) {
	    /* Compute the mean and std dev, and save them for later. */
	    imageBMean =
	      bruteCalcMatchWindowMean(imageB,
				       matchWindowTop, matchWindowBottom,
				       matchWindowLeft, matchWindowRight);
	    imageBStdDev = 
	      bruteCalcMatchWindowStdDev(imageB, imageBMean, 
					 matchWindowTop, matchWindowBottom,
					 matchWindowLeft, matchWindowRight);
	    STORE_MEAN(searchI, searchJ, imageBMean);
	    STORE_DEV(searchI, searchJ, imageBStdDev);
	  }

	  /* Compute Pearson's r to determine to corrolation between the
	     match window in image A and the current match window in the
	     search window in image B. */
	  
	  runningSum = 0;
	  for (matchJ=matchWindowTop;matchJ<=matchWindowBottom; matchJ++) {
	    for (matchI=matchWindowLeft;matchI<=matchWindowRight; matchI++) {

	      xdiff = matchI - searchI;
	      ydiff = matchJ - searchJ;

	      
	      runningSum += ((I(imageA, (imageI + xdiff), (imageJ + ydiff))
			      - imageAMean) *
			     (I(imageB, matchI, matchJ) - imageBMean));
	    }
	  }
	  currentR = (runningSum / ((N - 1) * imageAStdDev * imageBStdDev));
	  
	  /* If this location is the best yet, save it's disparity. */
	  if ((1 - currentR) < (1 - minR)) {
	    minR = currentR;
	    dx = (searchI - imageI);
	    dy = (searchJ - imageJ);
	  }
	} /* End iteration over search window column. */
      } /* End iteration over search window row. */
      
      /* We've determined the dx and dy disparities that result in the
	 minimum values for Pearson's r.  Now we should save them. */
      imRef(dxDisparityImage, imageI, imageJ) = (short)dx;
      imRef(dyDisparityImage, imageI, imageJ) = (short)dy;
    } /* Close loop over column in image. */
  } /* Close loop over row in image. */

#ifdef REPORT
  printf("Completed computation of disparities.  Exiting normCorrelation.\n");
#endif
  
  /* Clean up the images that we used. */
  imFree(meanMatrix);
  imFree(stdDevMatrix);
} /* Close _normCorrelation_ */


/* bruteCalcMatchWindowMean
 *
 *    input: image (either A or B) and boundaries of match window
 *    output: floating point mean of the intensities in the match window
 */
float bruteCalcMatchWindowMean(GrayImage image,
			       int matchWindowTop, int matchWindowBottom,
			       int matchWindowLeft, int matchWindowRight) {
  /* Declare variables local to _bruteCalcMatchWindowMean_. */
  int matchJ, matchI, N;
  float matchWindowMean, matchWindowSum;
  
  /* Compute number of pixels in match window. */
  N = ((matchWindowRight - matchWindowLeft + 1) *
       (matchWindowBottom - matchWindowTop + 1));
  matchWindowSum = 0;
  for (matchJ = matchWindowTop; matchJ <= matchWindowBottom; matchJ++) {
    for (matchI = matchWindowLeft; matchI <= matchWindowRight; matchI++) {
      matchWindowSum += I(image, matchI, matchJ);
    }
  }
  matchWindowMean = (matchWindowSum / N);
  return(matchWindowMean);
}

/* bruteCalcMatchWindowStdDev
 *
 *    input: image (either A or B) and boundaries of match window
 *    output: floating point standard deviation of the intensities in the
 *            match window
 */
float bruteCalcMatchWindowStdDev(GrayImage image,
				 float matchWindowMean,
				 int matchWindowTop, int matchWindowBottom,
                                 int matchWindowLeft, int matchWindowRight) {
  /* Declare variables local to _bruteCalcMatchWindowStdDev_. */
  int matchJ, matchI, N;
  float matchWindowStdDev, matchWindowSum;

  N = ((matchWindowRight - matchWindowLeft + 1) *
       (matchWindowBottom - matchWindowTop + 1));
  
  matchWindowSum = 0;
  for (matchJ = matchWindowTop; matchJ <= matchWindowBottom; matchJ++) {
    for (matchI = matchWindowLeft; matchI <= matchWindowRight; matchI++) {
      matchWindowSum += SQR(I(image, matchI, matchJ) - matchWindowMean);
    }
  }
  matchWindowStdDev = sqrt(matchWindowSum/(N - 1));

  return(matchWindowStdDev);
}

/* writeDisparityImageFile
 *
 * input: an image, filename, and information about the parameters that
 *        created the disparity image
 * output: void
 *
 * Writes the image to the file named by the given filename.
 */

void writeDisparityImageFile(ShortImage image, char *filename,
		    int hDisparity, int vDisparity,
                    int hNeighborhood, int vNeighborhood) {
  char *commentString;

  commentString = (char *)malloc(2000*sizeof(char));
  sprintf(commentString, "hDisp = %s\nvDisp = %s\nhNeigh = %s\nvNeigh = %s",
	  hDisparity, vDisparity, hNeighborhood, vNeighborhood);
  imSetComment(image, commentString);
  
  if (imSave(image, filename) == -1) {
    printf("Error writing image file! Aborting.\n");
    exit(1);
  }
}

/* rankTransform
 *
 *     input: dimensions of the input images
 *     output: void
 *
 *    Computes horizontal and vertical motion disparity images using
 *    normalized correlation.  Function is executed for effect on global
 *    output images.
 **/
void rankTransform(int imageWidth, int imageHeight) {
  /* Declaration of variables local to _rankTransform_. */
  GrayImage rankTransformedImageA, rankTransformedImageB;
  ShortImage absDiffMatrix;
  int imageI, imageJ, searchI, searchJ, matchI, matchJ;
  int hOffset, vOffset;
  int croppedHeight, croppedWidth;
  int rankHeight, rankWidth;
  int croppedImageTop, croppedImageBottom, croppedImageLeft, croppedImageRight;
  int matchWindowTop, matchWindowBottom, matchWindowLeft, matchWindowRight;
  int searchWindowTop, searchWindowBottom, searchWindowLeft, searchWindowRight;
  short dx, dy, runningSum;
  int xdiff, ydiff;
  int minL;
  
  /* Figure out the boundaries of the "cropped image," the area of the
     image for which motion disparities are computable. */
  hOffset = (hMaxDisparity + hConvNeighborhood + hRankNeighborhood);
  vOffset = (vMaxDisparity + vConvNeighborhood + vRankNeighborhood);
  croppedHeight = (imageHeight - (2 * hOffset));
  croppedWidth = (imageWidth - (2 * vOffset));
  croppedImageTop = vOffset;
  croppedImageBottom = (vOffset + croppedHeight - 1);
  croppedImageLeft = hOffset;
  croppedImageRight = (hOffset + croppedWidth - 1);
  
  /* Create matix images to store pixel ranks as they are dynamically computed.
     Set the origin of the image to correspond to the cropped area of the
     image. Do a sanity check. */
  rankHeight = (imageHeight - (2 * vRankNeighborhood));
  rankWidth = (imageWidth - (2 * hRankNeighborhood));

  rankTransformedImageA =
    (GrayImage)imNew(IMAGE_GRAY, rankWidth, rankHeight);
  rankTransformedImageB =
    (GrayImage)imNew(IMAGE_GRAY, rankWidth, rankHeight);

  imSetOffset(rankTransformedImageA, hRankNeighborhood, vRankNeighborhood);
  imSetOffset(rankTransformedImageB, hRankNeighborhood, vRankNeighborhood);
  
  if ((rankTransformedImageA == (GrayImage)NULL) ||
      (rankTransformedImageB == (GrayImage)NULL)) {
    printf("Could not create rankTransformedImage! Aborting.\n");
    exit(1);
  }

  /* Perform rank transformations. */
  rankTransformedImageA = createRankTransformedImage(imageA, imageWidth,
						     imageHeight);
  rankTransformedImageB = createRankTransformedImage(imageB, imageWidth,
						     imageHeight);

  /* Create a matrix to store absolute difference values for the search
     window. */
  absDiffMatrix = (ShortImage)imNew(IMAGE_SHORT,
                                    (2 * hMaxDisparity + 1),
                                    (2 * vMaxDisparity + 1));
  
  if (absDiffMatrix == (ShortImage)NULL) {
    printf("Could not create absDiffMatrix! Aborting.");
    exit(1);
  }

  for (imageJ = vOffset; imageJ < (imageHeight - vOffset); imageJ++) {
#ifdef REPORT
    printf("Evaluating disparities for pixels in row %d.\n", imageJ);
#endif
    for (imageI = hOffset; imageI < (imageWidth - hOffset); imageI++) {

      /* Initialize minimum L_1 value and dx, dy disparities. */
      minL = INT_MAX;
      dx = 0;
      dy = 0;
      
      /* Find the boundaries of the match window in image A. */
      matchWindowTop = (imageJ - vConvNeighborhood);
      matchWindowBottom = (imageJ + vConvNeighborhood);
      matchWindowLeft = (imageI - hConvNeighborhood);
      matchWindowRight = (imageI + hConvNeighborhood);
      
      /* Find the boundaries of the search window for pixel
	 (imageI, imageJ) and setup the matrix. */
      searchWindowTop = (imageJ - vMaxDisparity);
      searchWindowBottom = (imageJ + vMaxDisparity);
      searchWindowLeft = (imageI - hMaxDisparity);
      searchWindowRight = (imageI + hMaxDisparity);
      
      imSetOffset(absDiffMatrix, searchWindowLeft, searchWindowTop);
      /* Compute the sum of absolute differences for the UL pixel by
	 brute force. */
      searchI = searchWindowLeft;
      searchJ = searchWindowTop;
      
      /* Iterate over the search window in image B. */
      for (searchJ=searchWindowTop;searchJ<= searchWindowBottom; searchJ++) {
	for (searchI=searchWindowLeft;searchI<= searchWindowRight; searchI++) {
	  /* Find the boundaries of the match window in image B for
	     the current location in the search window. */
	  matchWindowTop = (searchJ - vConvNeighborhood);
	  matchWindowBottom = (searchJ + vConvNeighborhood);
	  matchWindowLeft = (searchI - hConvNeighborhood);
	  matchWindowRight = (searchI + hConvNeighborhood);
	  /* Compute pixels in top row of search window by brute force. */
	  if (searchJ == searchWindowTop) {
	    /* Compute the pixels in the top row by brute force. */
	    runningSum = 0;
	    for (matchJ=matchWindowTop;matchJ<=matchWindowBottom; matchJ++) {
	      for (matchI=matchWindowLeft;matchI<matchWindowRight; matchI++) {
		xdiff = matchI - searchI;
		ydiff = matchJ - searchJ;
		
		/* Compute the sum of absolute differences. */
		runningSum += abs(I(rankTransformedImageA,
				    (imageI + xdiff), (imageJ + ydiff)) -
				  I(rankTransformedImageB, matchI, matchJ));
	      }
	    }
	  }
	  
	  /* Use neighbor's absolute difference to make computation
	     easier for all the rest of the pixels in the image. */
	  else {
	    /* Get the sum of absolute differences value for the match
	       window neighboring one pixel above. */
	    runningSum = imRef(absDiffMatrix, searchI, searchJ - 1);
 	    /* Remove the absolute differences for that row. */
	    matchJ = (matchWindowTop - 1);
	    
	    for (matchI = matchWindowLeft;matchI<=matchWindowRight; matchI++) {
	      xdiff = matchI - searchI;
	      ydiff = matchJ - searchJ + 1;
	      
	      runningSum -= abs(I(rankTransformedImageA,
				  (imageI + xdiff), (imageJ + ydiff)) -
				I(rankTransformedImageB, matchI, matchJ));
	    }
	    /* Add the absolute differences for the row that's newly added
	       to the match window. */
	    matchJ = matchWindowBottom;
	    for (matchI = matchWindowLeft;matchI<=matchWindowRight; matchI++) {
	      xdiff = matchI - searchI;
	      ydiff = matchJ - searchJ + 1;
	      
	      runningSum += abs(I(rankTransformedImageA,
				  (imageI + xdiff), (imageJ + ydiff)) -
				I(rankTransformedImageB, matchI, matchJ));
	    }
	  }
	  /* If it's the best we've seen, save the disparity. */
	  if (runningSum < minL) {
	    minL = runningSum;
	    dx = (searchI - imageI);
	    dy = (searchJ - imageJ);
	  }
	  /* Store this absolute difference to make computation easier
	     for neighboring pixels. */
	  imRef(absDiffMatrix, searchI, searchJ) = runningSum;
	} /* Close loop over column in search window. */
      } /* Close loop over row in search window. */
      
      /* We've determined the dx and dy disparities that result in the
	 minimum values for Pearson's r.  Now we should save them. */
      imRef(dxDisparityImage, imageI, imageJ) = (short)dx;
      imRef(dyDisparityImage, imageI, imageJ) = (short)dy;
    } /* Close loop over column in image. */
  } /* Close loop over row in image. */
  
#ifdef REPORT
  printf("Completed computation of disparities.  Exiting _rankTransform_.\n");
#endif
  
  /* Clean up the images that we used. */
  imFree(rankTransformedImageA);
  imFree(rankTransformedImageB);
  imFree(absDiffMatrix);
} /* Close _rankTransform_ */


/* createRankTransformedImage
 *
 *    input: input image and its dimensions
 *    ouput: rank transformed image
 *
 */
GrayImage createRankTransformedImage(GrayImage image, int imageWidth,
				     int imageHeight) {
  
  /* Declare variables that are local to _createrankTransformedImage_. */
  GrayImage transformedImage;
  int croppedWidth, croppedHeight;
  int hOffset, vOffset;
  int rankWindowLeft, rankWindowRight, rankWindowTop, rankWindowBottom;
  int imageI, imageJ, rankI, rankJ;
  int rankCounter;
  float rankPixelIntensity, neighborsPixelIntensity;
  char *filename;
  
  /* We can only compute the rank transformed image for a cropped area,
     to ensure that the rank neighborhood is completely within the image. */
  hOffset = hRankNeighborhood;
  vOffset = vRankNeighborhood;
  croppedWidth = imageWidth - (2 * hOffset);
  croppedHeight = imageHeight - (2 * vOffset);
  
  /* Create an empty image, sets its origin to the proper offset, and do a
     sanity check. */
  transformedImage =
    (GrayImage)imNew(IMAGE_GRAY, croppedWidth, croppedHeight);
  imSetOffset(transformedImage, hOffset, vOffset);
  if (transformedImage == (GrayImage)NULL) {
    printf("Could not create transformedImage! Aborting.\n");
    exit(1);
  }

  /* Iterate over all the computable pixels in the image. */
  
  for (imageJ = (vOffset + 1); imageJ < (vOffset + croppedHeight); imageJ++) {
    for (imageI = hOffset; imageI < (hOffset + croppedWidth); imageI++) {
      rankWindowLeft = (imageI - hRankNeighborhood);
      rankWindowRight = (imageI + hRankNeighborhood);
      rankWindowTop = (imageJ - vRankNeighborhood);
      rankWindowBottom = (imageJ + vRankNeighborhood);
      rankPixelIntensity = I(image, imageI, imageJ);
      rankCounter = 0;
      
      /* For all the pixels in the rank neighborhood, count the number
	 of pixels with leser intensity. */
      for (rankJ = rankWindowTop; rankJ <= rankWindowBottom; rankJ++) {
	for (rankI = rankWindowLeft; rankI <= rankWindowRight; rankI++) {
	  if (I(image, rankI, rankJ) < rankPixelIntensity) {
	    rankCounter++;
	  }
	}
      }
      /* Store the rank for this pixel. */
      imRef(transformedImage, imageI, imageJ) = rankCounter;
    } /* Close loop over rows in the image. */
  } /* Close loop over columns in the image. */
  
  /* We're done, get outta here. */
  return(transformedImage);
}
