
#include "image.h"
#include <math.h>
#include <limits.h>
#include <stdio.h>

#define USE_MACROS

#ifdef IMAGE_CHECKALL
static char rcsid[] = "<snakes> encountered an image check failure.";
#endif

/* SQR(i) = (i)*(i) */
#define SQR(x) ((x) * (x))

/* Returns the x or y location in the image corresponding to location p in
   the match window. Assumes that floor(a/b) == a/b if a and b are integers. */
#define xoff(x, p) (x + p%w - w/2)
#define yoff(y, p) (y + p/w - w/2)

/* The two arrays in struct layer actually represent matrices. index(a, b) gives the
   correct index into the array corresponding to matrix cell (a, b), assuming that
   the matrix is represented in row-major order */
#define index(row, column) (numPositions*row + column)

/* External energy Eext of point (pixel) at position c in the local (square matrix)
   search space. Returns a huge energy if that pixel is outside the bounds of the
   image.*/
#define Eext(point, c) ((xoff(point.x, c) < 0 || \
			 yoff(point.y, c) < 0 || \
			 xoff(point.x, c) >= imageWidth || \
			 yoff(point.y, c) >= imageHeight) \
			? FLT_MAX \
			: -(imRef(externalEnergyImage, xoff(point.x, c), yoff(point.y, c))))

     /* Compute finite differences approximation to first derivative at vertex V[i] when:
   vertex V[i] is in position c
   vertex V[i+1] is in position n.

   Also take care of the special case that arise because the last vertex is
   connected to the first vertex. */
#define DERIV1TERM(i, c, n) \
     ( \
     (i < numVertices - 1) \
     ? \
     (SQR(xoff(V[i+1].x, n) - xoff(V[i].x, c)) + SQR(yoff(V[i+1].y, n) - yoff(V[i].y, c))) \
     : \
     (SQR(xoff(V[0].x, n) - xoff(V[i].x, c)) + SQR(yoff(V[0].y, n) - yoff(V[i].y, c))) \
     )

/* Compute finite differences approximation to the second derivative at vertex V[i] when:
   vertex V[i-1] is in position p,
   vertex V[i] is in position c,
   vertex V[i+1] is in position n 

   Also take care of the special cases that arise because the last vertex is
   connected to the first vertex. */
#define DERIV2TERM(i, c, p, n) \
     ( \
(i == 0) \
       ? \
       (SQR(xoff(V[1].x, n) - 2*xoff(V[0].x, c) + xoff(V[numVertices-1].x, p)) + \
	SQR(yoff(V[1].y, n) - 2*yoff(V[0].y, c) + yoff(V[numVertices-1].y, p))) \
       : \
       (\
(i < numVertices - 1) \
	? \
     (SQR(xoff(V[i+1].x, n) - 2*xoff(V[i].x, c) + xoff(V[i-1].x, p)) + \
      SQR(yoff(V[i+1].y, n) - 2*yoff(V[i].y, c) + yoff(V[i-1].y, p))) \
     : \
     (SQR(xoff(V[0].x, n) - 2*xoff(V[i].x, c) + xoff(V[i-1].x, p)) + \
      SQR(yoff(V[0].y, n) - 2*yoff(V[i].y, c) + yoff(V[i-1].y, p))) \
      ) \
      ) 
	   
/* structures for snakes */
typedef struct {
  int x;
  int y;
} point;

typedef struct {
  float *E;      /* Matrix of energies for this vertex. Each row represents
		    vertex V[i] in one of its numPositions possible positions.
		    Each column represents V[i+1] in one of its numPositions
		    possible positions. The matrix is represented as an array.*/
  int *prevVp;   /* The matrix of positions of the previous vertex
		    corresponding to the values of the matrix E. */
} layer;  

/* Prototypes */
void generateExternalEnergies(void);
void snakes(point V[], int numVertices, int numPositions, float alpha, float beta);
int gradientMag(FloatImage img, FloatImage *mag);
void createOutputImageFile(GrayImage im, char *filename,
			   point V[], int numVertices,
			   float alpha, float beta,
			   int numPositions);

/* Global Variables */
GrayImage inputImage;
FloatImage externalEnergyImage;    

char *inputFileName;
char *outputFileName;
char *externalEnergyFileName = "NONE GIVEN";
char *pointsFileName;
int imageWidth, imageHeight;
int outputRate = 0;
int gradientFlag;

/*
  V[]          = Initial points on the snake
  numVertices  = The number of initial points on the snake
  numPositions = Number of possible positions each vertex should search (hence
                 the search window is cube of width sqrt(numPositions)).
		 numPositions must be odd.
  alpha        = Amount by which to weight the first order term	    
  beta         = Amount by which to weight the second order term	        
*/
void snakes(point V[], int numVertices, int numPositions, float alpha, float beta) {  

  layer *L; 
  int i, delta;
  int p, c, n; /* positions of the previous, current, and next vertex */
  int w;
  float minimumCost;
  float temp;
  float minE;
  int positionOfMin = -1;
  int tempn;
  float oldMinE = FLT_MAX; /* minimum Energy for the snake from iteration n-1*/
  float oldOldMinE = FLT_MAX; /* minimum Energy for the snake from iteration n-2*/
  float dE = FLT_MAX;      /* change in energy between cycles */
  int bestn, bestc;
  int iteration = 0;
  char frameName[60];
  
  /* NUMPOSITIONS must be odd and a perfect square*/
  if ((numPositions % 2 == 0) || (trunc(sqrt(numPositions)) != sqrt(numPositions))) {
    printf("numPositions must be an odd perfect square. Aborting.\n");
    exit(1);
  }

  /* Make sure the all the vertex points inside the image. */
  for (i=0; i<numVertices; i++) { 
    if ((V[i].x < 0) || (V[i].x >= imageWidth) ||
	(V[i].y < 0) || (V[i].y >= imageHeight)) {
      printf("All initial snake points must be within the image. Aborting.\n");
      exit(1);      
    }
  }
  
  /* Calculate width of search window */
  w = sqrt(numPositions);
  printf("%d possible positions for each vertex\nSearch window is %d x %d\n", numPositions, w, w);


  /* Allocate and initialize one layer for each snake vertex (except the last one) */
  L = (layer *)malloc((numVertices) * sizeof(layer)); 
  for (i=0; i<numVertices; i++) {
    L[i].E = (float *)malloc(numPositions*numPositions * sizeof(float));
    L[i].prevVp = (int *)malloc(numPositions*numPositions * sizeof(int));
  }
  
  createOutputImageFile(inputImage, "test_before.pgm", V, numVertices, alpha, beta, numPositions);
  
  printf("\nInitialization finished. Starting minimization:\n\n");
  
  /* Keep iterating the minimization until the snake finds the minimum energy in its area */
  while (dE != 0) {

    printf("Iteration %d: Minimizing Vertices\n", iteration);
    
  /* Cycle through each vertex (except the last), one at a
     time, calculating the minimum energy of the snake at
     each of the numPositions possible positions of vertex V[i] */

    for (i=0; i<numVertices; i++) {  /* For each vertex V[i] */
            
      for (c=0; c<numPositions; c++) {   /* For each possible position of this vertex */
	
	for (n=0; n<numPositions; n++) { /* For each possible position of the next vertex */
	  
	  /* First fill in the external energy of the point
	     V[i] at its (x,y) position in the image */
	  L[i].E[index(c, n)] = Eext(V[i], c);
	  	  
	  /* Add in the cost from the first order term */
	  L[i].E[index(c, n)] += alpha * DERIV1TERM(i, c, n);
	  
	  /*
	    If we're not at the first vertex, we have to consider costs from:
	    1) second derivative term
	    2) minimum energy of the previous vertex, depending on its position 
	    
	    The plan of attack is thus: 
	    
	    Loop over all possible positions of the previous vertex, and figure out
	    which position of the previous vertex minimizes both the energy of the
	    previous vertex AND the second derivative at the current vertex. 
	    
	    Also store (in V[i].prevVp[index(c,n)]) the position of the previous
	    vertex which facilitated that minimum cost, so that after we minimize all
	    the snake vertex energies we can walk backward through the layers and figure out
	    the correct positions of every vertex to minimize the energy of the entire
	    snake. */
	  
	  minimumCost = FLT_MAX; /* set min cost to some huge value */
	  
	  for (p=0; p<numPositions; p++) { /* For all possible positions of previous vertex */
	    
	    if (i>0) {
	      temp = L[i-1].E[index(p, c)];  /* Energy when the vertex V[i-1] is at position p
						and the vertex V[i] is at position c */
	    }
	    else {
	      /* Special case for first vertex. Estimate the total energy as the
		 external energy, since we don't have enough info at this point to
		 construct the total energy (it's a circular process...) */
	      temp = Eext(V[i], p);
	    }
	    
	    if (beta != 0) {	      
	      
	      /* Add the cost of the second derivative when vertex
		 V[i] is in position c, V[i-1] is in position p,
		 and V[i+1] is in position n */
	      temp += beta * DERIV2TERM(i, c, p, n); 
	      
	    }
	    
	    if (temp < minimumCost) {
	      minimumCost = temp;           /* Set the new minimum cost. */
	      L[i].prevVp[index(c, n)] = p; /* Store the position of the previous vertex. */
	    }
	    
	  } /* end loop over p */
	  
	  /* Store minimum cost from second order term */
	  L[i].E[index(c, n)] += minimumCost;
	  
	} /* end loop over n */
	
      } /* end loop over c */
      
    } /* end loop over i*/
    
    printf("Tracing Backwards to get new vertex positions...\n");
    
    /* Now we walk backwards through the layers to extract out the positions
       of the snake vertices which minimize the energy of the entire snake. 
       
       First scan the top layer and find the cell with the minimum energy value.
       This cell position (a, b) tells us the position of vertices V[n-1] (the
       last vertex).
       
       Then we look up the value of L[n-1].prevVp[index(a, b)], which
       contains the position of vertex V[n-2] that minimized the energy of
       vertices V[n-1].

       We do this for each vertex, so that in the end we have the positions of all
       of the vertices which minimizes the energy of the entire snake. */
    
    minE = FLT_MAX;
    
    for (c = 0; c<numPositions; c++) {
      for (n = 0; n<numPositions; n++) {
	if (L[numVertices-1].E[index(c, n)] < minE) {
	  
	  minE = L[numVertices-1].E[index(c, n)];
	  positionOfMin = L[numVertices-1].prevVp[index(c, n)];
	  bestn = n;
	  bestc = c;

	}
      }
    }

    tempn = bestc;

    V[numVertices-1].x = xoff(V[numVertices-1].x, bestc);
    V[numVertices-1].y = yoff(V[numVertices-1].y, bestc);
    
    positionOfMin = L[numVertices-1].prevVp[index(bestc, bestn)];
    
    for (i = numVertices-2; i>=0; i--) {
      
      V[i].x = xoff(V[i].x, positionOfMin);
      V[i].y = yoff(V[i].y, positionOfMin);
      
      if (i !=0 ) {
	/* remember the position of V[i] */
	tempn = positionOfMin;
	
	/* Figure out the position of vertex V[i-1] */
	positionOfMin = L[i].prevVp[index(positionOfMin, tempn)];
      }
    }


    dE = fabs(oldMinE - minE);
    printf("(oldOldMinE = %g oldMinE = %g newMinE = %g\nChange in Energy = %g\n\n",
	   oldOldMinE, oldMinE, minE, dE);

    /* Stop iterating if the snake is waffling back and forth between
       two configurations. */
    if ((oldOldMinE == minE) && (iteration != 0))
      break;

    oldOldMinE = oldMinE; /* remember the minE from the previous iteration */
    oldMinE = minE;       /* remember the minE from this iteration */

    /* Render a frame if necessary */
    if ((outputRate > 0) && (iteration % outputRate == 0)) {
      sprintf(frameName, "frame%d.pgm", iteration/outputRate);
      createOutputImageFile(inputImage, frameName, V, numVertices, alpha, beta, numPositions);
    }

    iteration++;
  }
  
}

int main(int argc, char **argv) {
  
  int numPoints;      /* number of snake vertices */
  point *snakePoints; /* the vertices */
  float alpha, beta;  /* snake parameters (weights for 1st and 2nd order terms) */
  int numPositions;
  
  /* Make sure we have enough input arguments. */
  if (argc < 7) {
    printf("Too few arguments given to program. Aborting.\n");
    exit(1);
  }

  /* Read in the arguments. */
  pointsFileName = argv[1];
  inputFileName = argv[2];
  outputFileName = argv[3];
  numPositions = atoi(argv[4]);
  alpha = atof(argv[5]);
  beta = atof(argv[6]);
  outputRate = atoi(argv[7]);
  gradientFlag = atoi(argv[8]);
  externalEnergyFileName = argv[9];
    
  /* Attempt to load in the input image. */
  inputImage = (GrayImage)imLoad(IMAGE_GRAY, inputFileName);
  if (inputImage == (GrayImage)NULL) {
    printf("Error loading input image! Aborting.\n");
    exit(1);
  }

  /* Set the dimensions of the images */
  imageWidth = (int)imGetWidth(inputImage); 
  imageHeight = (int)imGetHeight(inputImage);
  
  /* Print debugging info about the input images */
  printf("inputImage = %ux%u\n", imageWidth, imageHeight);

  /* Load/Generate external energy for all pixels */
  generateExternalEnergies();

  /* print out the rendering frame rate */
  if (outputRate > 0) {
    printf("Outputting an image every %d iterations\n", outputRate);
  }
  else {
    outputRate = 0;
    printf("Outputting only initial and final frames.\n");
  }
  
  /* Read in the snake points */
  numPoints = readSnakePoints(&snakePoints);

  printf("\n\n");
  
  /* Minimize the snake completely */
  snakes(snakePoints, numPoints, numPositions, alpha, beta); 
			  
  /* write the output image */
  createOutputImageFile(inputImage, outputFileName, snakePoints, numPoints, alpha, beta, numPositions);

  imFree(inputImage);
  imFree(externalEnergyImage);
  
  return(0);
}

/**********************************************************************************/


/* Calculates the external energy Eext(i,j) for all (i, j) and
 places them in the global FloatImage externalEnergyImage */
void generateExternalEnergies(void) {
  
  int l;
  FloatImage temp;

  if (externalEnergyFileName == NULL) {
    /* Use input image as base for external energy */

    printf("Using %s as external energy image...\n", inputFileName);
    if ((temp = (FloatImage)imConvert(inputImage, IMAGE_FLOAT)) == NULL) {
      printf("Couldn't convert %s to a FloatImage. Aborting.\n", inputFileName);
      exit(1);
    }
  }
  else {

    /* Use the external energy image that the user specified, converting it to
       the correct type if necessary */
    
    printf("Using %s as external energy image...\n", externalEnergyFileName);
    l = strlen(externalEnergyFileName);

    if (externalEnergyFileName[l-2] == 'f') {
      temp = (FloatImage)imLoad(IMAGE_FLOAT, externalEnergyFileName);
      if (temp == NULL) {
	printf("Couldn't load %s. Aborting.\n", externalEnergyFileName);
	exit(1);
      }
    }
    else {
      /* Convert it to a FloatImage */
      
      if (externalEnergyFileName[l-2] == 's') {
	/* We have to convert it from a .psm to a .pfm */
	ShortImage im;
	
	printf("Converting %s to a FloatImage...\n", externalEnergyFileName);
	
	im = (ShortImage)imLoad(IMAGE_SHORT, externalEnergyFileName);
	temp = (FloatImage)imConvert(im, IMAGE_FLOAT);
	imFree(im);
      }
      else if (externalEnergyFileName[l-2] == 'd') {
	/* We have to convert it from a .pdm to a .pfm */
	DoubleImage im;
	
	printf("Converting %s to a FloatImage...\n", externalEnergyFileName);
	
	im = (DoubleImage)imLoad(IMAGE_DOUBLE, externalEnergyFileName);
	temp = (FloatImage)imConvert(im, IMAGE_FLOAT);
	imFree(im);
      }
      else {
	printf("%s must be a FloatImage (.pfm), ShortImage (.psm) or DoubleImage (.pdm). Aborting.\n",
	       externalEnergyFileName);
	exit(1);
      }
      
      if (temp == NULL) {
	printf("Couldn't convert %s to a FloatImage. Aborting.\n", externalEnergyFileName);
	exit(1);
      }
    } /* end -- converting to a FloatImage */

  } /* end -- loading specified external energy image */


  if (gradientFlag != 0) {
    /* Replace the external energy data with its squared magnitude gradient */
    printf("Computing the squared magnitude gradient of the external energy.\n");
    if (gradientMag(temp, &externalEnergyImage) != 0) {
      printf("Error in gradientMag(). Aborting.\n");
      exit(1);
    }
  }
  else {
    /* Use the external energy unmodified */
    externalEnergyImage = temp;
  }

}

/* Read in the initial positions of the snake vertices from a file.
   Return the number of vertices. */
int readSnakePoints(point **V) {

  FILE *pfile;
  int x, y;
  int count = 0;
  int i = 0;
  
  /* yes, I know this is a slimey hack. */

  if ((pfile = fopen(pointsFileName, "r")) == NULL) {
    printf("Error opening snake points file. Aborting.\n");
    exit(1);
  }

  while (fscanf(pfile, "%d %d", &x, &y) != EOF)
    count ++;

  printf("Read %d snake points\n", count);
  
  fclose(pfile);

  if ((*V = (point *)malloc(count * sizeof(point))) == NULL) {
    printf("Couldn't allocate memory for snake points. Aborting.\n");
    exit(1);
  }
  
  if ((pfile = fopen(pointsFileName, "r")) == NULL) {
    printf("Error opening snake points file. Aborting.\n");
    exit(1);
  }

  while (fscanf(pfile, "%d %d", &x, &y) != EOF) {
    printf("x=%d y=%d\n", x, y);
    (*V)[i].x = x;
    (*V)[i].y = y;
    i++;
  }

  fclose(pfile);

  return count;
  
}

/* This bresenham line-drawing code snarfed from Graphics Gems Vol. 1 */
void drawLine(GrayImage im, int x1, int y1, int x2, int y2) {
  int d, x, y, ax, ay, sx, sy, dx, dy;
  
#define C(i, j) imRef(im, i, j)
#define cFlip(i, j) C(i, j) = (256 - C(i, j))
#define Flip(i, j) C(i, j) = 0
#define SGN(a) (((a)<0) ? -1 : 1) 

  dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
  dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);
  
  x = x1;
  y = y1;
  if (ax>ay) {                /* x dominant */
    d = ay-(ax>>1);
    for (;;) {
      Flip(x, y);
      if (x==x2) return;
      if (d>=0) {
	y += sy;
	d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else {                      /* y dominant */
    d = ax-(ay>>1);
    for (;;) {
      Flip(x, y);
      if (y==y2) return;
      if (d>=0) {
	x += sx;
	d -= ay;
      }
      y += sy;
      d += ax;
    }
  }
}


void createOutputImageFile(GrayImage im, char *filename,
			   point V[], int numVertices,
			   float alpha, float beta, int numPositions) {
  
  int error;
  char *commentString;
  GrayImage copy;
  int i;
  
  copy = (GrayImage)imDup(im);
  if (copy == (GrayImage)NULL) {
    printf("Couldn't duplicate image in createOutputImageFile(). Aborting.\n");
    exit(1);
  }
  
  commentString = (char *)malloc(2000*sizeof(char));
  sprintf(commentString, "this file = %s\ninputImage = %s\noutputImage = %s\nexternalEnergyImage = %s\nnumPositions = %d\nalpha = %g\nbeta = %g\n",
	  filename, inputFileName, outputFileName, externalEnergyFileName,
	  numPositions, alpha, beta);

  imSetComment(copy, commentString);
 
  for (i = 0; i<numVertices-1; i++) {
    drawLine(copy, V[i].x, V[i].y, V[i+1].x, V[i+1].y);
  }
  drawLine(copy, V[numVertices-1].x, V[numVertices-1].y, V[0].x, V[0].y);

  if (imSave(copy, filename) == -1) {
    printf("Error saving image in createOutputImageFile(). Aborting.\n");
    exit(1);
  }

  imFree(copy);
  
 }


/*
  I stole this code from lib-support in the robotlab :
   
   Computes the magnitude of the gradient. Storage
   will be allocated for MAG if it is a pointer to a NULL
   FloatImage; otherwise the storage provided will be used. A non-zero value
   is returned on error -- e.g. on error allocating storage or if the result
   images passed are not the same size as the input image. */

int gradientMag(FloatImage img, FloatImage *mag) {
  
#define imGetStore(i)   ({ &(imRef((i), imGetXBase(i), imGetYBase(i))); })

  register float *imgp, *magp, *img_xendp;
  register float x, y;
  register long width;
  register float a, b, c, d;
  register float *img_endp;
  register long height;
  int made_mag = 0;
  
  
  if (img == (FloatImage) NULL  ||  mag == (FloatImage *) NULL)
    return -1;
  
  width  = imGetWidth(img);
  height = imGetHeight(img);
  
  if (*mag == (FloatImage) NULL) {
    if ((*mag = imNew(IMAGE_FLOAT, width, height)) == (FloatImage) NULL)
      goto gradient_error;
    
    made_mag = 1;
  }
  else if (imGetWidth(*mag) != width  ||  imGetHeight(*mag) != height)
    goto gradient_error;

  imgp = imGetStore(img);
  magp = imGetStore(*mag);
  
  for (img_endp = imgp + width*(height-1) - 1; imgp < img_endp; ) {
    a = *imgp;
    c = *(imgp+width);
    
    for (img_xendp = imgp + width - 1; imgp < img_xendp; ) {
      imgp++;

      b = *(imgp);
      d = *(imgp+width);

      a = d - a;
      c = b - c;
      x = a + c;
      y = a - c;

      a = b;
      c = d;
      
      *(magp++) = sqrt(x*x + y*y);
    }

    imgp++;				     /* last column gets 0.0 */
    *(magp++) = 0.0;
  }

  for (img_endp += width; imgp <= img_endp; imgp++) /* last row gets 0.0 */
    *(magp++) = 0.0;

  return 0;
  
 gradient_error:
  if (made_mag) imFree(*mag);
  return -1;
}


